from cas.common.models import BuildEnvironment, BuildSubsystem
from cas.common.config import DataResolverScope

from pathlib import Path

import importlib
import logging


class Sequencer:
    """
    Class that executes a number of discrete programs (subsystems) in order.
    """

    def __init__(self, path: Path, config: dict):
        self.env = BuildEnvironment(path, config)

        self._args = self.env.config.args
        self._subsystems = {}
        self._logger = logging.getLogger(__name__)

    def _load_subsystem(self, name: str, module: str, config: dict) -> BuildSubsystem:
        subsystem = self._subsystems.get(name)
        if subsystem is not None:
            return

        mod = importlib.import_module(module)
        if mod is None:
            raise Exception(f"Failed to load subsystem {mod}")
        self._logger.debug(f"loaded '{module}' subsystem")

        subsystem = mod._subsystem(self.env, config)
        self._subsystems[name] = subsystem

    def _run_subsystem(self, scope: DataResolverScope, name: str) -> bool:
        if self._args.dry_run:
            return True

        # get the unresolved configuration first to run checks
        subsystem = self.env.config.subsystems.get(name)
        if not subsystem:
            return True

        build_types = subsystem.get("build_types")
        if build_types and self.env.build_type not in build_types:
            self._logger.debug(f"subsystem {name} skipped (build type mismatch)")
            return True

        categories = subsystem.get("categories")
        if (
            self.env.build_categories
            and categories
            and len(self.env.build_categories.intersection(set(categories))) == 0
        ):
            self._logger.debug(f"subsystem {name} skipped (category mismatch)")
            return True

        # get the full configuration
        subsystem = subsystem.with_scope(scope)
        self._load_subsystem(name, subsystem.module, subsystem.get("options", {}))

        # run
        sys = self._subsystems[name]
        self._logger.info(f"running subsystem {name}")
        if self._args.clean:
            if not sys.clean():
                return False
        else:
            result = sys.build()
            scope._data.subsystems[name] = result
            if not result.success:
                return False
        return True

    def run(self) -> bool:
        # build whitelist/blacklist
        whitelist = self._args.include_subsystems
        blacklist = self._args.exclude_subsystems
        if whitelist:
            whitelist = whitelist.split(",")
        if blacklist:
            blacklist = blacklist.split(",")

        # create the scope
        scope = DataResolverScope()

        # run subsystems
        for sub in self.env.config.subsystems.keys():
            if whitelist and sub not in whitelist:
                self._logger.debug(f"subsystem {sub} skipped (not whitelisted)")
                continue
            if blacklist and sub in blacklist:
                self._logger.debug(f"subsystem {sub} skipped (blacklisted)")
                continue
            if not self._run_subsystem(scope, sub):
                return False
        return True



from cas.common.models import BuildResult, BuildSubsystem


class PanZipSubsystem(BuildSubsystem):
    def build(self) -> BuildResult:
        return BuildResult(True)

    def clean(self) -> bool:
        return True


_subsystem = PanZipSubsystem



from cas.common.models import BuildResult, BuildSubsystem
from pathlib import Path

import cas.common.utilities as utilities

import os
import shutil


def _shutil_delete_force(action, name, exc):
    os.chmod(name, os.path.stat.S_IWRITE)
    os.remove(name)


class SyncFolderSubsystem(BuildSubsystem):
    def build(self) -> BuildResult:
        from_dir = Path(self.config["from"]).resolve()
        to_dir = Path(self.config.to).resolve()

        if self.config.clobber and to_dir.exists():
            shutil.rmtree(to_dir, onerror=_shutil_delete_force)

        if not to_dir.exists():
            to_dir.mkdir()

        files = utilities.rglob_multi(from_dir, self.config.files)
        self._logger.debug(f"{len(files)} file(s) to copy")

        for src in files:
            src = src.resolve()
            dest = to_dir.joinpath(src.relative_to(from_dir))
            os.makedirs(dest.parent, exist_ok=True)
            shutil.copy(src, dest)

        return BuildResult(True)

    def clean(self) -> bool:
        to_dir = Path(self.config.to).resolve()
        if to_dir.exists():
            shutil.rmtree(to_dir, onerror=_shutil_delete_force)
        return True


_subsystem = SyncFolderSubsystem



import cas
from cas.common.config import DefaultValidatingDraft7Validator
from cas.common.models import BuildEnvironment, BuildResult, BuildSubsystem
from cas.common.cache import FileCache
from cas.common.assets.models import (
    Asset,
    AssetBuildContext,
    BaseDriver,
    SerialDriver,
    BatchedDriver,
)

import os
import json
import logging
import pathlib
import importlib
import multiprocessing

from typing import Mapping, Sequence, Callable, Any
from pathlib import Path

_schema_path = Path(cas.__file__).parent.absolute().joinpath("schemas")

logger = None


def _async_mod_init():
    global logger
    logger = multiprocessing.log_to_stderr(logging.INFO)


def _run_async_serial(
    context: AssetBuildContext, driver: SerialDriver, asset: Asset
) -> bool:
    relpath = os.path.relpath(asset.path, driver.env.root)

    context.logger = logger
    if not context.logger:
        context.logger = logging.getLogger(driver.__class__.__module__)

    context.logger.info(f"(CC) {str(relpath)}")
    success = driver.compile(context, asset)

    if not success:
        context.logger.error(f"  Failed compile {str(relpath)}")
    return success


def _run_async_batched(
    context: AssetBuildContext, driver: BatchedDriver, assets: Sequence[Asset]
) -> bool:
    context.logger = logger
    for asset in assets:
        relpath = os.path.relpath(asset.path, driver.env.root)
        context.logger.info(f"(CC) {str(relpath)}")
    return driver.compile_all(context, assets)


class AssetSubsystem(BuildSubsystem):
    def __init__(self, env: BuildEnvironment, config: Mapping[str, Any]):
        super().__init__(env, config)

        self._drivers = {}
        self._validators = {}

        self._args = self.env.config.args
        self._dry_run = self._args.dry_run

        self._cache = FileCache(self.env.cache, self.env.cache["assets"])

    def _get_asset_driver(self, name: str) -> BaseDriver:
        driver = self._drivers.get(name)
        if driver is not None:
            return driver

        mod = importlib.import_module(f"cas.common.assets.drivers.{name}")
        if mod is None:
            raise Exception(f"Invalid type {name}")
        self._logger.debug(f"loaded '{name}' driver")

        driver = mod._driver(self.env)
        self._drivers[name] = driver
        return driver

    def _load_asset_context(self, config: Mapping[str, Any]) -> AssetBuildContext:
        # validate the schema
        if config.type not in self._validators:
            driver_path = _schema_path.joinpath("drivers", f"{config.type}.json")
            if not driver_path.exists():
                raise Exception(
                    f"Unable to find schema for asset driver '{config.type}'"
                )
            with open(driver_path, "r") as f:
                self._validators[config.type] = DefaultValidatingDraft7Validator(
                    json.load(f)
                )
        if config.get("options") is not None:
            self._validators[config.type].validate(config.options._data)

        srcpath = Path(config.src)
        if not srcpath.exists():
            raise Exception(f'The asset source folder "{srcpath}" does not exist.')

        # find everything by the patterns
        patterns = []
        if isinstance(config.files, str):
            patterns.append(config.files)
        elif isinstance(config.files, Sequence):
            patterns += config.files
        else:
            raise NotImplementedError()

        files = []
        for pattern in patterns:
            for path in pathlib.Path(srcpath).absolute().rglob(pattern):
                if not path.is_file():
                    continue
                files.append(path)

        # create context and add assets
        context = AssetBuildContext(config)
        for f in files:
            context.assets.append(Asset(f, {}))
        return context

    def _build_assets(
        self,
        contexts: Sequence[AssetBuildContext],
        callback: Callable[[Callable[[Mapping[str, Any]], bool], Sequence[Any]], None],
    ):
        for context in contexts:
            if len(context.assets) <= 0:
                self._logger.warning(
                    f"no files found for a context with type {context.config.type}"
                )
                continue

            if isinstance(context.driver, BatchedDriver):
                callback(_run_async_batched, (context, context.driver, context.assets))
            elif isinstance(context.driver, SerialDriver):
                for asset in context.assets:
                    callback(_run_async_serial, (context, context.driver, asset))
            else:
                raise Exception("Unknown driver type")

    def _build_assets_sync(self, contexts: Sequence[AssetBuildContext]) -> bool:
        """
        Builds assets synchronously.
        """
        jobs = []

        _async_mod_init()
        def callback(func: Callable[[Mapping[str, Any]], bool], params: Sequence[Any]):
            jobs.append(func(*params))

        self._build_assets(contexts, callback)
        return all(job for job in jobs)

    def _build_assets_async(self, contexts: Sequence[AssetBuildContext]) -> bool:
        """
        Builds assets asynchronously.
        """
        jobs = []
        pool = multiprocessing.Pool(self._args.threads, initializer=_async_mod_init)

        def callback(func: Callable[[Mapping[str, Any]], bool], params: Sequence[Any]):
            jobs.append(pool.apply_async(func, params))

        try:
            self._build_assets(contexts, callback)
            pool.close()
        except KeyboardInterrupt:
            pool.terminate()
        pool.join()

        return all(job.get() for job in jobs)

    def _run_asset_build(self, clean: bool = False) -> bool:
        contexts = []
        for entry in self.config.assets:
            contexts.append(self._load_asset_context(entry))

        hash_inputs = {}
        hash_outputs = {}
        total_build = 0

        # prebuild
        for context in contexts:
            assets = context.assets.copy()

            context.assets = []
            context.driver = self._get_asset_driver(context.config.type)

            for asset in assets:
                result = context.driver.precompile(context, asset)
                if not result:
                    self._logger.error("Asset dependency error!")
                    return False

                # clean outputs if requested
                if clean is True:
                    for f in result.outputs:
                        if not f.exists():
                            continue
                        f.unlink()
                    continue

                # check hashes
                invalidated = False
                for f in result.inputs:
                    f = f.resolve()
                    if not os.path.exists(f):
                        self._logger.error(
                            f"Required dependency '{f}' could not be located!"
                        )
                        return False
                    if not self._cache.validate(f):
                        invalidated = True

                for f in result.outputs:
                    f = f.resolve()
                    if not self._cache.validate(f):
                        invalidated = True

                aid = asset.get_id()
                hash_inputs[aid] = result.inputs
                hash_outputs[aid] = result.outputs

                if invalidated:
                    total_build += 1
                    context.assets.append(asset)

        if clean is True:
            self._logger.info("assets cleaned")
            return True

        self._logger.info(
            f"{len(hash_inputs)} input files, {len(hash_outputs)} output files"
        )
        self._logger.info(f"{total_build} files total will be rebuilt")

        if self._dry_run or total_build == 0:
            return True

        # build
        threaded = self._args.threads > 1
        if threaded:
            self._logger.info(
                f"running multithreaded build with {self._args.threads} threads"
            )
        else:
            self._logger.info("running singlethreaded build")

        # sort out drivers that should run synchronously regardless of threading
        sync_assets = [
            context
            for context in contexts
            if not context.driver.threadable() or not threaded
        ]
        async_assets = [
            context for context in contexts if context.driver.threadable() and threaded
        ]

        if not self._build_assets_sync(sync_assets):
            self._logger.error("Build failed")
            return False
        if not self._build_assets_async(async_assets):
            self._logger.error("Build failed")
            return False

        self._logger.info("recalculating asset hashes...")
        for context in contexts:
            for asset in context.assets:
                # save updated hashes
                aid = asset.get_id()
                for f in hash_inputs[aid]:
                    self._cache.put(f)
                for f in hash_outputs[aid]:
                    self._cache.put(f)

        self.env.cache.save()
        return True

    def build(self) -> BuildResult:
        return BuildResult(self._run_asset_build())

    def clean(self) -> bool:
        return self._run_asset_build(True)


_subsystem = AssetSubsystem



from cas.common.models import BuildResult, BuildSubsystem, BuildEnvironment
from cas.common.buildsys.msbuild import MSBuildCompiler
from cas.common.buildsys.posix import PosixCompiler
from cas.common.buildsys.vpc import VPCInstance

import cas.common.utilities
import os
import sys
from pathlib import Path


class BuildsysSubsystem(BuildSubsystem):
    def __init__(self, env: BuildEnvironment, config: dict):
        super().__init__(env, config)

        # if a platform isn't specified,
        # default to the 64-bit version of our current platform
        self._platform = config.get("platform")
        if not self._platform:
            self._platform = cas.common.utilities.resolve_platform_name()

        if cas.common.utilities.is_platform_windows():
            self._compiler = MSBuildCompiler(self.env, self.config, self._platform)
        else:
            self._compiler = PosixCompiler(self.env, self.config, self._platform)

    def build(self) -> BuildResult:
        # configure stage (run VPC, build makefiles)
        if self.config.configure:
            # first we need to bootstrap VPC
            self._logger.info("bootstrapping VPC")
            if not self._compiler.bootstrap():
                return BuildResult(False)

            self._logger.info("running VPC")
            vpc = VPCInstance(self.env, self.config, self._platform)
            if not vpc.run() or not self._compiler.configure():
                return BuildResult(False)

        # compile stage (compile dependencies and engine)
        if self.config.compile:
            self._logger.info("compiling")
            if not self._compiler.build():
                return BuildResult(False)

        return BuildResult(True)

    def clean(self) -> bool:
        # clean output files before we delete project files!
        if not self._compiler.clean():
            self._logger.error("Output binary clean failed!")
            return False

        sln_ext = f"{self._platform}.sln"
        proj_ext = f"{self._platform}.vcxproj"
        for root, _, files in os.walk(self.env.src):
            for f in files:
                if (
                    f.endswith(".vpc_crc")
                    or f.endswith(".vpc_cache")
                    or f.endswith(sln_ext)
                    or f.endswith(proj_ext)
                ):
                    Path(root).joinpath(f).unlink()

        return True


_subsystem = BuildsysSubsystem



from cas.common.models import BuildEnvironment
from typing import Mapping

import logging


class BaseCompiler:
    """
    Base compiler class from which all compilers should extend.
    """

    def __init__(self, env: BuildEnvironment, config: Mapping, platform: str):
        self._env = env
        self._config = config
        self._platform = platform
        self._logger = logging.getLogger(self.__class__.__module__)

    def bootstrap(self) -> bool:
        """
        Compiles any dependencies required for the configure stage.
        """
        return True

    def clean(self) -> bool:
        """
        Removes all output files of the project.
        """
        raise NotImplementedError()

    def configure(self) -> bool:
        """
        Generates the necessary files to build the project.
        """
        return NotImplementedError()

    def build(self) -> bool:
        """
        Compiles the project.
        """
        raise NotImplementedError()



import os
import sys
import json
import hashlib
import struct
import logging

import tqdm

from pathlib import Path
from typing import List, Mapping, Any


class TqdmLoggingHandler(logging.Handler):
    def __init__(self, level=logging.NOTSET):
        super().__init__(level)

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.tqdm.write(msg)
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            self.handleError(record)


def relative_paths(root: Path, paths: list) -> List[str]:
    """
    Normalises paths from incoming configuration and ensures
    they are all strings relative to root
    """
    result = []
    for path in paths:
        # more hacks for exclusions I'm not happy about
        # maybe we should subclass Path to make this cleaner?
        exclusion = path.startswith("!")
        if exclusion:
            path = path[1:]

        # make sure paths are relative!
        if isinstance(path, Path):
            inp = str(path.relative_to(root))
        elif isinstance(path, str):
            inp = path
            if os.path.isabs(path):
                inp = os.path.relpath(path, root)
        else:
            raise NotImplementedError()

        if exclusion:
            inp = "!" + inp
        result.append(inp)
    return result


def rglob_invert(patterns: List[str]) -> List[str]:
    """
    Inverts a rglob condition.
    """
    result = []
    for pattern in patterns:
        if pattern.startswith("!"):
            result.append(pattern[1:])
        else:
            assert "!" not in pattern
            result.append("!" + pattern)
    return result


def rglob_multi(root: Path, patterns: List[str]) -> List[Path]:
    """
    Advanced recursive glob of a path collapsing multiple include/exclude patterns.
    """
    files = []
    for pattern in patterns:
        # patterns starting with ! are treated as exclusions
        exclusion = pattern.startswith("!")
        if exclusion:
            pattern = pattern[1:]

        for path in root.rglob(pattern):
            if not path.is_file():
                continue
            if exclusion:
                if path in files:
                    files.remove(path)
            else:
                if path in files:
                    continue
                files.append(path)

    return files


def paths_to_relative(root, paths) -> List:
    out = []
    for path in paths:
        out.append(os.path.relpath(path, root))
    return out


def hash_file_sha256(path: Path) -> str:
    hash = hashlib.sha256()
    with open(path, "rb") as f:
        while True:
            data = f.read(65536)
            if not data:
                break
            hash.update(data)
    return hash.hexdigest()


def hash_object_sha256(obj: Any) -> str:
    data = json.dumps(obj, sort_keys=True)
    hash = hashlib.sha256()
    hash.update(data.encode())

    return hash.hexdigest()


def resolve_platform_name() -> str:
    plat = sys.platform
    arch = struct.calcsize("P") * 8
    assert arch in [32, 64]

    if plat == "win32" or plat.startswith("msys") or plat.startswith("cygwin"):
        plat = "win"
    elif plat.startswith("freebsd"):
        plat = "freebsd"
    elif plat.startswith("openbsd"):
        plat = "openbsd"
    elif plat.startswith("darwin"):
        plat = "osx"
    elif plat.startswith("linux"):
        pass
    else:
        raise Exception(f"Unsupported platform {plat}")
    return f"{plat}{str(arch)}"


# Simply resolves the OS name, doesn't care about the actual arch
# Possible returns: win, osx, linux, freebsd, openbsd
def resolve_os_name() -> str:
    plat = sys.platform
    if plat == "win32" or plat.startswith("msys") or plat.startswith("cygwin"):
        plat = "win"
    elif plat.startswith("freebsd"):
        plat = "freebsd"
    elif plat.startswith("openbsd"):
        plat = "openbsd"
    elif plat.startswith("darwin"):
        plat = "osx"
    elif plat.startswith("linux"):
        plat = "linux"
    return plat


def is_platform_windows() -> bool:
    plat = sys.platform
    if plat == "win32" or plat.startswith("msys") or plat.startswith("cygwin"):
        return True
    return False


def is_platform_linux() -> bool:
    return sys.platform.startswith("linux")


def is_platform_osx() -> bool:
    return sys.platform.startswith("osx")


def get_dotpath_value(key: str, mapping: Mapping) -> Any:
    keys = key.split(".")
    current = mapping
    for k in keys:
        if k not in current:
            raise KeyError(k)
        current = current[k]
        if k != keys[-1] and not isinstance(current, Mapping):
            raise KeyError(k)
    return current


def set_dot_notation(target: dict, key: str, value):
    keys = key.split(".")
    pre = target
    pre_k = None
    last = target
    for key in keys:
        pre = last
        last = last[key]
        pre_k = key
    pre[pre_k] = value


def map_to_envvars(envvars: Mapping) -> Mapping[str, str]:
    result = {}
    for k, v in envvars.items():
        if isinstance(v, bool):
            result[k] = "1" if v else "0"
        else:
            result[k] = str(v)
    return result



import cas.common.utilities as utilities
from cas.common.steamtools import SteamInstance
from cas.common.config import ConfigurationUtilities
from cas.common.cache import CacheManager

from pathlib import Path
import os
import sys
import logging
import subprocess


class BuildEnvironment:
    """
    Contains attributes about the current build environment
    """

    def __init__(self, path: str, config: dict):
        self.config = ConfigurationUtilities.parse_root_config(path, config)

        self.cache = CacheManager(path)
        self.cache.load()

        self.build_type = self.config.args.build_type
        self.build_categories = None

        categories = self.config.args.build_categories
        if categories:
            categories = categories.split(",")
            self.build_categories = frozenset(categories)

        self.verbose = self.config.args.verbose

        self.root = self.config.path.root
        self.content = self.config.path.content
        self.game = self.config.path.game
        self.src = self.config.path.src

        self.platform = utilities.resolve_platform_name()
        self.steam = None

        self._setup_bindir()

    def _get_appid_folder(self, appid: int) -> Path:
        if not self.steam:
            self.steam = SteamInstance()
        for app in self.steam.apps:
            if app.appid == appid:
                return app.path
        return None

    def _check_autodetect_appid(self) -> bool:
        if self.src.exists():
            logging.debug("appid autodetect skipped - src dir exists")
            return False
        elif self.get_tool("chaos").exists():
            logging.debug("appid autodetect skipped - chaos executable exists")
            return False

        return True

    def _setup_bindir(self):
        self.bindir = self.game.joinpath("bin", self.platform)
        override = self.config.options.get("bin_path")

        # we have a bin path override
        if override is not None:
            self.bindir = Path(override).joinpath(self.platform)
        # we're a mod (modwrapper present but no game executable) - autodetect bin path from appid
        elif self._check_autodetect_appid():
            appid = self.config.options.get("bin_appid")
            if appid is not None:
                logging.debug(f"mod detected - using base appid {str(appid)}")
                self.bindir = self._get_appid_folder(appid).joinpath(
                    "bin", self.platform
                )

        if not self.bindir.exists():
            raise Exception("Could not find the bin directory")

        logging.debug(f"using bin directory {self.bindir}")

    """
    Retrieves the absolute path to the tool at the specified source path.
    If the source path is None, it will default to self.bindir.
    """

    def get_tool(self, tool: str, src: Path = None) -> Path:
        if src is None:
            src = self.bindir
        assert not tool.endswith(".exe")
        if utilities.is_platform_windows():
            tool += ".exe"
        return src.joinpath(tool).resolve()

    def get_lib(self, lib: str) -> Path:
        if utilities.is_platform_windows():
            lib += ".dll"
        elif utilities.is_platform_osx():
            lib += ".dylib"
        elif utilities.is_platform_linux():
            lib += ".so"
        else:
            raise NotImplementedError()
        return self.bindir.joinpath(lib).resolve()

    def run_subprocess(self, *args, **kwargs):
        predef = {}
        if not self.verbose:
            predef["stdout"] = subprocess.DEVNULL
            predef["stderr"] = subprocess.DEVNULL
        return subprocess.run(*args, **dict(predef, **kwargs))

    def run_tool(self, *args, **kwargs) -> int:
        predef = {}
        predef["env"] = os.environ
        predef["env"]["VPROJECT"] = str(self.config.path.vproject)

        try:
            result = self.run_subprocess(*args, **dict(predef, **kwargs))
        except Exception as e:
            raise Exception(f"failed to execute tool with parameters: {args}") from e
        return result.returncode


class BuildResult:
    """
    Represents a result returned from a subsystem
    """

    def __init__(self, success: bool, outputs: dict = {}):
        self.success = success
        self.outputs = outputs


class BuildSubsystem:
    """
    Represents a build system that implements custom behaviour.
    """

    def __init__(self, env: BuildEnvironment, config: dict):
        self.env = env
        self.config = config

        mod = self.__class__.__module__
        self._logger = logging.getLogger(mod)

    def build(self) -> BuildResult:
        """
        Invokes the build logic of the subsystem.
        Returns a BuildResult with the result.
        """
        raise NotImplementedError()

    def clean(self) -> bool:
        """
        Removes all the output files generated by this subsystem.
        """
        raise NotImplementedError()



import logging
import itertools
import cas.common.utilities as utilities
from cas.common.models import BuildEnvironment
from cas.common.config import LazyDynamicDotMap
from cas.common.cache import FileCache

import sys
from typing import List

BUILD_TYPE_MAP = {
    "trunk": "CHAOS_TRUNK_BUILD",
    "staging": "CHAOS_STAGING_BUILD",
    "release": "CHAOS_REL_BUILD",
}


class VPCArguments:
    def __init__(
        self,
        args: List[str],
        args_raw: List[str],
        groups: List[str],
        defines: List[str],
    ):
        self.args = args
        self.raw = args_raw
        self.groups = groups
        self.defines = defines

    def to_string(self) -> str:
        params = []
        for x in self.raw:
            params.append(x)
        for x in self.args:
            params.append(f"/{x}")
        for x in self.groups:
            params.append(f"+{x}")
        for x in self.defines:
            params.append(f"/define:{x}")
        return " ".join(params)

    def to_list(self) -> List[str]:
        params = []
        for x in self.raw:
            params.append(x)
        for x in self.args:
            params.append(f"/{x}")
        for x in self.groups:
            params.append(f"+{x}")
        for x in self.defines:
            params.append(f"/define:{x}")
        return params


class VPCInstance:
    def __init__(self, env: BuildEnvironment, config: LazyDynamicDotMap, platform: str):
        self._env = env
        self._config = config.vpc
        self._solution = config.solution
        self._group = config.group
        self._platform = platform

        self._cache = self._env.cache["vpc"]
        if "files" not in self._cache:
            self._cache["files"] = {}

        self._file_cache = FileCache(self._env.cache, self._cache["files"])
        self._logger = logging.getLogger(__name__)

    def _list_all_vpcs(self) -> list:
        return itertools.chain(
            self._env.src.rglob("*.vpc"), self._env.src.rglob("*.vgc")
        )

    def _process_vpc_args(self) -> VPCArguments:
        args = list(self._config.args)
        defines = list(self._config.defines)

        args.append(self._solution)
        args.append(self._platform)
        args.append(self._config.windows.toolchain)

        if self._config.ide_files:
            args.append("clangdb")
            args.append("cmake")

        raw = []
        raw.append("/mksln")
        raw.append(f"{self._solution}_{self._group}_{self._platform}")

        build_type = self._env.build_type
        defines.append(BUILD_TYPE_MAP[build_type])

        if build_type == "trunk":
            defines.append("DEVELOPMENT_ONLY")

        return VPCArguments(args, raw, [self._group], defines)

    def _clear_crc_files(self):
        # Don't do this on Windows, it's an unnecessary slowdown
        if utilities.is_platform_windows():
            return
        crc_files = self._env.src.rglob("*.vpc_crc")
        for f in crc_files:
            f.unlink()

    def _verify_argument_hash(self, args: List) -> bool:
        old_hash = self._cache.get("args", None)
        new_hash = utilities.hash_object_sha256(args)
        self._cache["args"] = new_hash

        if not old_hash:
            return False
        return old_hash == new_hash

    def run(self) -> bool:
        args = self._process_vpc_args()
        rebuild = False  # do we need to rebuild?

        # hash the arguments
        if not self._verify_argument_hash(args.to_list()):
            rebuild = True

        # hash the VPC files
        vpc_files = self._list_all_vpcs()
        for f in vpc_files:
            if not self._file_cache.validate(f):
                self._file_cache.put(f)
                rebuild = True

        if rebuild:
            self._file_cache.garbage_collect()
            self._env.cache.save()
            self._clear_crc_files()

        if not rebuild:
            self._logger.info("scripts are unchanged, not running VPC")
            return True

        # select the right executable for our platform
        if utilities.is_platform_windows():
            vpc_bin = "vpc"
        elif utilities.is_platform_osx():
            vpc_bin = "vpc_osx"
        elif utilities.is_platform_linux():
            vpc_bin = "vpc"
        else:
            raise NotImplementedError()

        args = [
            self._env.get_tool(vpc_bin, self._env.config.path.devtools.joinpath("bin"))
        ] + args.to_list()
        ret = self._env.run_tool(args, cwd=self._env.src)

        # ensure cache is invalidated if vpc fails
        if not ret == 0:
            self._env.cache.vpc = {}
            self._env.cache.save()
            return False
        return True



from cas.common.models import BuildResult, BuildSubsystem
from pathlib import Path

import cas.common.utilities
import sys
import getpass


class SteamworksSubsystem(BuildSubsystem):
    """
    Subsystem that pushes app depots to Steamworks using the content builder.
    App and depot configurations are generated on the fly
    """

    def _get_credentials(self):
        user = self.config.username
        pwd = self.config.password
        if not user or not pwd:
            print("Enter Steamworks Credentials")
            if not user:
                user = input("Username: ")
            if not pwd:
                pwd = getpass.getpass(prompt="Password: ")
        return user, pwd

    def _run_steamcmd(self) -> bool:
        tool_dir = Path(self.config.tooldir).resolve()
        tool_path = None

        if cas.common.utilities.is_platform_windows():
            tool_path = tool_dir.joinpath("builder", "steamcmd.exe")
        elif cas.common.utilities.is_platform_osx():
            tool_path = tool_dir.joinpath("builder_osx", "steamcmd.sh")
        elif cas.common.utilities.is_platform_linux():
            tool_path = tool_dir.joinpath("builder_linux", "steamcmd.sh")
        else:
            raise NotImplementedError(f"unsupported platform {sys.platform}")

        script_cmd = []
        for script in self.config.scripts:
            script_file = tool_dir.joinpath("scripts", f"app_build_{script}.vdf")
            if not script_file.exists():
                self._logger.error(
                    f'Unable to find SteamCMD script at "{script_file}"!'
                )
                return False
            script_cmd.append("+run_app_build_http")
            script_cmd.append(script_file)

        user, pwd = self._get_credentials()
        args = [tool_path, "+login", user, pwd] + script_cmd + ["+quit"]
        ret = self.env.run_tool(args, cwd=self.env.src)
        return ret == 0

    def build(self) -> BuildResult:
        return BuildResult(self._run_steamcmd())

    def clean(self) -> bool:
        return True


_subsystem = SteamworksSubsystem



from pathlib import Path
import sys
import logging
import cas.common.utilities

import vdf

if cas.common.utilities.is_platform_windows():
    import winreg


def get_steam_path() -> Path:
    if cas.common.utilities.is_platform_windows():
        steam_key = winreg.OpenKey(
            winreg.HKEY_LOCAL_MACHINE,
            r"SOFTWARE\WOW6432Node\Valve\Steam",
            0,
            winreg.KEY_READ,
        )
        steam_path = winreg.QueryValueEx(steam_key, "InstallPath")
        winreg.CloseKey(steam_key)

        return Path(steam_path[0]).resolve()
    else:
        raise NotImplementedError(
            "TODO: implement platforms other than Windows for autodetection!"
        )


class SteamApp:
    """
    Represents a singular Steam application.
    """

    @staticmethod
    def from_acf(library: Path, path: Path):
        with open(path, "r") as f:
            parsed = vdf.load(f)
        state = parsed["AppState"]

        result = SteamApp()
        result.appid = int(state["appid"])
        result.name = state.get("name")
        result.path = library.joinpath("common", state["installdir"])
        return result


class SteamInstance:
    """
    Represents an instance of Steam.
    """

    def __init__(self):
        self._path = get_steam_path()
        self._load_apps()

    def _load_apps(self):
        root_apps = self._path.joinpath("steamapps")
        fpath = root_apps.joinpath("libraryfolders.vdf")
        if not fpath.exists():
            raise Exception("Unable to find libraryfolders.vdf!")
        with open(fpath, "r") as f:
            lib_vdf = vdf.load(f)

        logging.debug("libraryfolders.vdf found")

        libraries = []
        folders = lib_vdf["LibraryFolders"]
        for k, v in folders.items():
            if not k.isdigit():
                continue
            libraries.append(Path(v).joinpath("steamapps").resolve())

        libraries.append(root_apps.resolve())
        logging.debug(f"{len(libraries)} num libraries detected")

        apps = []
        for l in libraries:
            for path in l.glob("appmanifest_*.acf"):
                if not path.is_file():
                    continue
                app = SteamApp.from_acf(l, path)
                if not app.path.exists():
                    logging.warn(
                        f"Skipping steam app {app.appid} as ACF reported as installed but we cannot locate the folder"
                    )
                    continue
                apps.append(app)
        self.apps = apps



from cas.common.assets.models import (
    Asset,
    AssetBuildContext,
    SerialDriver,
    PrecompileResult,
)

from typing import List, Set
from pathlib import Path
import os


class ModelDriver(SerialDriver):
    """
    Driver that handles compiling studio model files
    with either mdlcompile or studiomdl
    """

    def _tool_name(self):
        return "mdlcompile"

    def _deps_searchline(self, token, line, set):
        line = line.strip()
        i = line.find(token)
        if i == -1:
            return
        s1 = line.rfind('"', 0, i) + 1
        s2 = line.find('"', i)

        path = line[s1:s2]
        set.append(path)

    def _parse_deps_from_vdf(self, path: Path):
        """
        Walks a QC/MC VDF file and extracts the input and output paths.
        """
        inputs = []
        outputs = []

        # parse the qc/mc and extract dependent paths
        with open(str(path), "r") as f:
            qc = f.readlines()

        for line in qc:
            # exclude comments and specific strings
            if line.startswith("//"):
                continue
            if "$includemodel" in line:
                continue
            self._deps_searchline("smd", line, inputs)
            self._deps_searchline("mdl", line, outputs)

        return set(inputs), set(outputs)

    def _convert_relpaths(self, root: Path, paths: Set[Path]) -> Set[Path]:
        result = set()
        for path in paths:
            result.add(root.joinpath(path))
        return result

    def precompile(self, context: AssetBuildContext, asset: Asset) -> List[str]:
        # TODO: perhaps we should have another way of doing this rather than jankily "autodetecting" the dest
        gamedir = os.path.relpath(asset.path.parent, self.env.root)
        gamedir = gamedir.replace("\\", "/").split("/")[1]

        inputs, outputs = self._parse_deps_from_vdf(asset.path)
        inputs = self._convert_relpaths(asset.path.parent, inputs)
        outputs = self._convert_relpaths(
            self.env.game.joinpath(gamedir, "models"), outputs
        )

        extra = set()
        for f in outputs:
            # also add a .dx90.vtx and .vvd for every .mdl
            if f.suffix == ".mdl":
                extra.add(f.with_suffix(".dx90.vtx"))
                extra.add(f.with_suffix(".vvd"))

        inputs.add(asset.path)
        return PrecompileResult(inputs, outputs | extra)

    def compile(self, context: AssetBuildContext, asset: Asset) -> bool:
        args = [str(self.tool), str(asset.path)]

        returncode = self.env.run_tool(args)
        return returncode == 0


_driver = ModelDriver



from cas.common.models import BuildEnvironment
from cas.common.buildsys.shared import BaseCompiler
import cas.common.utilities as utilities

import os
import sys
from typing import List, Dict

winreg = None
if utilities.is_platform_windows():
    import winreg

# JM: todo: this should probably not be hardcoded
MSBUILD_PATH = "C:/Program Files (x86)/Microsoft Visual Studio/2019/Community/MSBuild/Current/Bin/amd64/MSBuild.exe"


class MSBuildCompiler(BaseCompiler):
    """
    MSBuild compiler, used on Windows
    """

    def __init__(self, env: BuildEnvironment, config: dict, platform: str):
        super().__init__(env, config.windows, platform)
        self._solution = f"{config.solution}_{config.group}_{platform}.sln"
        self._project = config.get("project")

        self._build_type = config.type
        self._setup_winsdk()

    def _setup_winsdk(self):
        sdk_key = winreg.OpenKey(
            winreg.HKEY_LOCAL_MACHINE,
            r"SOFTWARE\WOW6432Node\Microsoft\Microsoft SDKs\Windows\v10.0",
            0,
            winreg.KEY_READ,
        )
        sdk_path = winreg.QueryValueEx(sdk_key, "InstallationFolder")
        sdk_ver = winreg.QueryValueEx(sdk_key, "ProductVersion")

        winreg.CloseKey(sdk_key)
        self._winsdk_path = os.path.join(sdk_path[0], f"bin\\{sdk_ver[0]}.0\\x64")

    def _invoke_msbuild(self, targets: List[str], parameters: Dict[str, str]) -> bool:
        args = [MSBUILD_PATH, self._solution]
        args.append("-target:" + ";".join(targets))
        for k, v in parameters.items():
            args.append(f"/p:{k}={v}")

        self._logger.debug(f"Running MSBuild with parameters: {args}")
        returncode = self._env.run_tool(args, cwd=self._env.src)
        return returncode == 0

    def _build_default_parameters(self) -> Dict[str, str]:
        params = {}
        if self._config.type == "debug":
            params["Configuration"] = "Debug"
        else:
            params["Configuration"] = "Release"
            params["DebugSymbols"] = "false"
            params["DebugType"] = "None"
        return params

    def _build_internal(self, target: str) -> bool:
        params = self._build_default_parameters()
        if self._project:
            target = f"{self._project}:{target}"
        return self._invoke_msbuild([target], params)

    def clean(self) -> bool:
        return self._build_internal("Clean")

    def configure(self) -> bool:
        # msbuild doesn't need to do anything
        return True

    def build(self) -> bool:
        return self._build_internal("Build")



from cas.common.assets.models import (
    Asset,
    AssetBuildContext,
    SerialDriver,
    PrecompileResult,
)

from typing import List


class CaptionDriver(SerialDriver):
    """
    Driver that handles compiling closed captions
    """

    def _tool_name(self):
        return "captioncompiler"

    def precompile(self, context: AssetBuildContext, asset: Asset) -> List[str]:
        asset.outpath = asset.path.with_suffix(".dat")
        return PrecompileResult([asset.path], [asset.outpath])

    def compile(self, context: AssetBuildContext, asset: Asset) -> bool:
        args = [str(self.tool), str(asset.path)]

        returncode = self.env.run_tool(args)
        return returncode == 0


_driver = CaptionDriver



import cas.common.utilities as utilities

import json
import os
from pathlib import Path
from typing import Mapping

from dotmap import DotMap


class CacheManager:
    def __init__(self, root: Path):
        self._root = root
        self._file = root.joinpath("content", ".cas_cache.json")
        self._caches = DotMap()

    def load(self):
        if os.path.exists(self._file):
            with open(self._file, "r") as f:
                self._caches = DotMap(json.loads(f.read()))

    def save(self):
        with open(self._file, "w") as f:
            f.write(json.dumps(self._caches.toDict()))

    def __getitem__(self, key):
        if key not in self._caches:
            self._caches[key] = {}
        return self._caches[key]

    def __setitem__(self, key, value):
        self._caches[key] = value


class FileCache:
    """
    Implements a cache of file hashes
    """

    def __init__(self, manager: CacheManager, cache: Mapping):
        self._manager = manager
        self._cache = cache

    def garbage_collect(self):
        # garbage collect paths that no longer exist
        self._cache = {
            k: v
            for k, v in self._cache.items()
            if self._manager._root.joinpath(k).exists()
        }

    def validate(self, path: Path):
        if not path.exists():
            return False
        rel = path.relative_to(self._manager._root)
        hash = self._cache.get(str(rel))
        if not hash:
            return False
        return hash == utilities.hash_file_sha256(path)

    def put(self, path: Path):
        if not path.exists():
            raise Exception(
                f'Tried to insert the path "{str(path)}" into the cache, which does not exist!'
            )
        rel = path.relative_to(self._manager._root)
        self._cache[str(rel)] = utilities.hash_file_sha256(path)



import cas.common.utilities as utilities
from cas.common.models import BuildEnvironment
from cas.common.buildsys.shared import BaseCompiler

import os
import shutil
import logging
import requests
import subprocess
import multiprocessing
from typing import List, Mapping
from pathlib import Path

import tqdm
import appdirs

DOCKER_IMAGE_TAG = "chaos-cas-steamrt:latest"
STEAMRT_REPO_URL = "https://repo.steampowered.com/steamrt-images-soldier/snapshots/latest-container-runtime-depot"
STEAMRT_IMAGE_URL = (
    STEAMRT_REPO_URL
    + "/com.valvesoftware.SteamRuntime.Sdk-amd64,i386-soldier-sysroot.tar.gz"
)
STEAMRT_IMAGE_SHA256 = (
    "4a77fbd1ba45286eedf0446ecfd7d42ab0cc53bc1bc0cecd53efd75ba040598f"
)
STEAMRT_HASH_URL = STEAMRT_REPO_URL + "/SHA256SUMS"


class BaseCompileEnvironment:
    """
    Base compilation environment
    """

    def __init__(self, env: BuildEnvironment, config: Mapping):
        self._env = env
        self._config = config
        self._env_config = config.environment.config
        self._env_vars = self._build_static_env()
        self._logger = logging.getLogger(self.__class__.__module__)

    def _build_static_env(self) -> dict:
        return {
            "CC": self._config.cc,
            "CXX": self._config.cxx,
            "JOBS": self._config.get("jobs", multiprocessing.cpu_count()),
            "PLATFORM": self._config.platform,
        }

    def _build_env(self, env: Mapping[str, str]) -> Mapping[str, str]:
        nenv = os.environ.copy()
        nenv.update(self._env_vars)
        nenv.update(env)

        return nenv

    def run(
        self, args: List[str], env: Mapping[str, str] = {}, path_suffix: str = None
    ) -> int:
        return NotImplementedError()


class NativeCompileEnvironment(BaseCompileEnvironment):
    """
    Compilation environment that executes commands in a native shell
    """

    def run(
        self, args: List[str], env: Mapping[str, str] = {}, path_suffix: str = None
    ) -> int:
        root_path = self._env.config.path.root
        cwd = str(root_path.joinpath("src"))
        if path_suffix:
            cwd += f"/{path_suffix}"
        print(cwd)
        return self._env.run_subprocess(
            args, env=self._build_env(env), cwd=cwd
        ).returncode


class ChrootCompileEnvironment(BaseCompileEnvironment):
    """
    Compilation environment that executes commands using schroot
    """

    def run(
        self, args: List[str], env: Mapping[str, str] = {}, path_suffix: str = None
    ) -> int:
        cwd = self._env.config.path.src.joinpath(path_suffix)
        defargs = ["schroot", "-c", self._env_config.name, "--", "bash", "-c"]

        return self._env.run_subprocess(
            defargs.extend(args), env=self._build_env(env), cwd=cwd
        ).returncode


class DockerCompileEnvironment(BaseCompileEnvironment):
    """
    Compilation environment that executes commands using Docker
    """

    def _ensure_installed(self) -> bool:
        cache_folder = Path(appdirs.user_cache_dir("chaos_cas")).joinpath("docker")
        cache_folder.mkdir(parents=True, exist_ok=True)

        # check docker to see if the image already exists
        rebuild_image = False
        ret = self._env.run_subprocess(
            ["docker", "inspect", "--type=image", DOCKER_IMAGE_TAG],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        if ret.returncode != 0:
            rebuild_image = True

        sysroot_file = cache_folder.joinpath("steamrt-sysroot.tar.gz")
        if not sysroot_file.exists():
            self._logger.info("starting download of SteamRT docker image")
            response = requests.get(STEAMRT_IMAGE_URL, stream=True, timeout=1)
            with tqdm.tqdm(
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
                total=int(response.headers["Content-Length"]),
                desc="Downloading SteamRT docker image",
            ) as progress:
                with open(sysroot_file, "wb") as f:
                    for chunk in response.iter_content(chunk_size=1024):
                        if chunk:
                            progress.update(len(chunk))
                            f.write(chunk)

            # Obtain the hash from the steam repo
            response = requests.get(STEAMRT_HASH_URL, stream=False, timeout=1)
            dockerfile_hash = STEAMRT_IMAGE_SHA256
            for line in response.iter_lines():
                line = line.decode("utf-8")
                if (
                    "com.valvesoftware.SteamRuntime.Sdk-amd64,i386-soldier-sysroot.tar.gz"
                    in line
                ):
                    s = line.split(" ")
                    dockerfile_hash = s[0]
                    break

            # verify the hash
            sysroot_hash = utilities.hash_file_sha256(sysroot_file)
            if not sysroot_hash == dockerfile_hash:
                self._logger.error("Failed to verify the SteamRT docker image hash")
                self._logger.error(f"Wanted {dockerfile_hash}")
                self._logger.error(f"Got {sysroot_hash}")
                sysroot_file.unlink()
                return False

            # ensure we rebuild the image as our file changed
            rebuild_image = True

        if rebuild_image:
            shutil.copyfile(
                Path(__file__).parent.joinpath("chaos-cas-steamrt.Dockerfile"),
                cache_folder.joinpath("Dockerfile"),
            )
            ret = self._env.run_subprocess(
                ["docker", "build", "--tag", DOCKER_IMAGE_TAG, "."], cwd=cache_folder
            )
            if not ret.returncode == 0:
                self._logger.error("Failed to build the SteamRT docker image")
                return False

        return True

    def run(
        self, args: List[str], env: Mapping[str, str] = {}, path_suffix: str = None
    ) -> int:
        if not self._ensure_installed():
            return 1

        root_path = self._env.config.path.root

        cwd = str(root_path.joinpath("src"))
        if path_suffix:
            cwd += f"/{path_suffix}"

        defargs = [
            "docker",
            "run",
            "--rm",
            "-i",
            "-u",
            f"{os.getuid()}:{os.getgid()}",
            "-v",
            f"{root_path}:{root_path}",
            "-w",
            cwd,
        ]

        nenv = self._env_vars.copy()
        nenv.update(env)

        for k, v in nenv.items():
            defargs.append("-e")
            defargs.append(f"{k}={v}")
        defargs.append(DOCKER_IMAGE_TAG)
        defargs.extend(args)
        self._logger.debug(defargs)

        return self._env.run_subprocess(defargs).returncode


class BaseDependency:
    """
    Represents an external third-party dependency that is not managed through VPC to build
    """

    def __init__(self, name: str, config: Mapping):
        self._name = name
        self._config = config

    def build(
        self, env: BaseCompileEnvironment, envvars: Mapping[str, str] = {}
    ) -> bool:
        return env.run(self._config.run, envvars, self._config.src) == 0


_compile_environments = {
    "native": NativeCompileEnvironment,
    "chroot": ChrootCompileEnvironment,
    "docker": DockerCompileEnvironment,
}


class PosixCompiler(BaseCompiler):
    """
    Generic POSIX compiler, used to build via generated makefiles
    """

    def __init__(self, env: BuildEnvironment, config: dict, platform: str):
        super().__init__(env, config.posix, platform)
        self._makefile = f"{config.solution}_{config.group}_{platform}.mak"
        self._project = config.get("project")

        self._build_type = config.type
        self._compile_env = _compile_environments[self._config.environment.type](
            env, self._config
        )
        self._dependencies = {
            BaseDependency(k, v) for k, v in self._config.dependencies.items()
        }

    def _build_dependencies(self, clean: bool = False) -> bool:
        bstr = "cleaning" if clean else "building"
        envvars = utilities.map_to_envvars({"CLEAN": clean})

        for dependency in self._dependencies:
            self._logger.info(f"{bstr} dependency {dependency._name}")
            if not dependency.build(self._compile_env, envvars):
                return False
        return True

    def _build_vpc(self, clean: bool = False) -> bool:
        args = ["clean"] if clean else []
        return self._run_makefile("Makefile", args, "utils/vpc")

    def _run_makefile(self, file: str, args: List[str], path_suffix: None) -> bool:
        jobs = self._config.get("jobs", multiprocessing.cpu_count())
        sanitizers = self._config.sanitizers

        args = [
            "make",
            "-f",
            file,
            f"-j{jobs}",
        ] + args

        envvars = {
            "CFG": self._build_type,
            "ASAN": sanitizers.address,
            "UBSAN": sanitizers.behavior,
            "TSAN": sanitizers.threading,
            "NO_STRIP": False,
            "NO_DBG_INFO": False,
            "VALVE_NO_AUTO_P4": True,
        }

        return (
            self._compile_env.run(args, utilities.map_to_envvars(envvars), path_suffix)
            == 0
        )

    def _run_project(self, clean: bool = False) -> bool:
        args = []
        if clean:
            args.append("clean")
        elif self._project is not None:
            args.append(self._project)

        return self._run_makefile(self._makefile, args, clean)

    def bootstrap(self) -> bool:
        return True

    def clean(self) -> bool:
        if not self._build_vpc(True):
            self._logger.error("vpc clean failed")

        if not self._build_dependencies(True):
            self._logger.error("dependency clean failed")

        if not self._run_project(True):
            self._logger.error("main project clean failed")

        return True

    def configure(self) -> bool:
        if not super().configure():
            return False

        if not self._build_dependencies():
            self._logger.error("dependency build failed; run with --verbose to see why")
            return False

        return True

    def build(self) -> bool:
        if not self._run_project():
            self._logger.error("build failed")
            return False
        return True



from cas.common.models import BuildResult, BuildSubsystem
from pathlib import Path

import sys
import shutil
import importlib.util
import contextlib


class FGDBuildSubsystem(BuildSubsystem):
    """
    Subsystem that builds FGDs and Hammer assets from HammerAddons
    """

    def build(self) -> BuildResult:
        project = self.env.config.options.project

        srcpath = Path(self.env.root).joinpath(self.config.source).resolve()
        destpath = Path(self.env.root).joinpath(self.config.dest).resolve()

        spec = importlib.util.spec_from_file_location(
            "hammeraddons.unifyfgd", srcpath.joinpath("unify_fgd.py")
        )
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)

        build_dest = srcpath.joinpath("build")
        if not build_dest.exists():
            build_dest.mkdir()

        outfile = build_dest.joinpath(f"{project}.fgd")

        # redirect our output so logs aren't spammed on non-verbose mode
        log_dev = sys.stdout if self.env.verbose else None
        with contextlib.redirect_stdout(log_dev):
            mod.action_export(
                srcpath.joinpath("fgd"),
                None,
                frozenset({"SRCTOOLS", self.config.branch}),
                outfile,
                False,
                False,
            )

        hammer_dir = destpath.joinpath("hammer")

        shutil.copy(outfile, hammer_dir.joinpath("cfg", f"{project}.fgd"))
        
        self.override_folder(srcpath.joinpath("hammer", "materials"), hammer_dir.joinpath("materials"))
        self.override_folder(srcpath.joinpath("hammer", "models"), hammer_dir.joinpath("models"))
        self.override_folder(srcpath.joinpath("hammer", "scripts"), hammer_dir.joinpath("scripts"))

        return BuildResult(True)

    def clean(self) -> bool:
        return True

    def override_folder(self, src, dest):
        if dest.exists():
            shutil.rmtree(dest)
        shutil.copytree(src, dest)


_subsystem = FGDBuildSubsystem



from cas.common.models import BuildEnvironment

import uuid
from pathlib import Path
from typing import List, Set


class PrecompileResult:
    def __init__(self, inputs: Set[Path], outputs: Set[Path]):
        self.inputs = inputs
        self.outputs = outputs


class AssetBuildContext:
    """
    A collection of assets with shared configuration
    """

    def __init__(self, config: dict):
        self.assets = []
        self.config = config


class Asset:
    """
    Represents an asset to be compiled
    """

    def __init__(self, path: Path, config: dict):
        self.id = uuid.uuid4()
        self.path = path
        self.config = config

    def get_id(self):
        """
        Returns the unique identifier for this asset
        """
        return self.id


class BaseDriver:
    """
    Represents an instance of a tool that compiles assets
    """

    def __init__(self, env: BuildEnvironment):
        self.env = env
        self.tool = str(self.env.get_tool(self._tool_name()))

    def _tool_name(self):
        raise NotImplementedError()

    def threadable(self) -> bool:
        """
        Whether this driver is capable of running asynchronously or not
        """
        return True

    def precompile(self, context: AssetBuildContext, asset: Asset) -> PrecompileResult:
        """
        Checks to ensure all required files are present
        Returns a list of source and output files to be hashed, or None if failure
        """
        raise NotImplementedError()


class SerialDriver(BaseDriver):
    def compile(self, context: AssetBuildContext, asset: Asset) -> bool:
        """
        Performs the compile
        Returns True if success, otherwise False.
        """
        raise NotImplementedError()


class BatchedDriver(BaseDriver):
    def compile_all(self, context: AssetBuildContext, assets: List[Asset]) -> bool:
        """
        Performs the compile
        Returns True if success, otherwise False.
        """
        raise NotImplementedError()



import cas
import cas.common.utilities as utilities

import sys
import json
import multiprocessing
from collections.abc import Sequence, Mapping
from pathlib import Path

import simpleeval
import jsonschema
from dotmap import DotMap


def extend_validator_with_default(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):
        for property, subschema in properties.items():
            if "default" in subschema:
                instance.setdefault(property, subschema["default"])
        for error in validate_properties(validator, properties, instance, schema):
            yield error

    return jsonschema.validators.extend(validator_class, {"properties": set_defaults})


DefaultValidatingDraft7Validator = extend_validator_with_default(
    jsonschema.Draft7Validator
)


class DataResolverScope(Mapping):
    def __init__(self):
        self._data = DotMap()

    def __getitem__(self, key):
        if key not in self._data:
            raise KeyError(key)
        return self._data[key]

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def get_subsystem_output(self, subsystem: str, output: str):
        return self._data.subsystems[subsystem].outputs[output]

    def exclude_subsystem_input_files(self, subsystem: str, root: Path):
        files = self.get_subsystem_output(subsystem, "files")
        return utilities.rglob_invert(utilities.relative_paths(root, files))


class DataResolver:
    """
    Class that actually does the configuration resolution
    """

    def __init__(self, data: Mapping):
        self._data = data

    def build_eval_locals(self, parent: Mapping, scope: DataResolverScope) -> Mapping:
        return DotMap(
            {
                "parent": parent,
                "context": scope,
                "path": self._data.path,
                "args": self._data.args,
                "assets": self._data.assets,
                "subsystems": self._data.subsystems,
                "env": {
                    "platform": sys.platform,
                    "cpu_count": multiprocessing.cpu_count(),
                },
            }
        )

    def _inject_config_str(self, config: str, eval_locals: Mapping) -> str:
        """
        A terrible lexical parser for interpolated globals.
        I.e. "build type: $(args.build)" returns "build type: trunk"
        """
        prev = None
        inblock = False
        current = ""
        result = ""

        for c in config:
            if c == "$":
                prev = c
                continue
            if not inblock and c == "(" and prev == "$":
                # read to end for key
                inblock = True
            elif inblock and c == ")":
                value = utilities.get_dotpath_value(current, eval_locals)
                if value is None:
                    raise Exception(
                        f"Value of configuration variable $({current}) was None"
                    )
                result += str(value)

                current = ""
                inblock = False
            elif inblock:
                current += c
            else:
                result += c
            prev = c
        return result

    def eval(self, condition: str, parent: Mapping, scope: DataResolverScope) -> bool:
        # avoid infinite recusion
        if isinstance(parent, LazyDynamicMapping):
            parent = parent._data

        eval_locals = self.build_eval_locals(parent, scope)

        injected = self._inject_config_str(condition, eval_locals)
        evaluator = simpleeval.EvalWithCompoundTypes(names=eval_locals)

        result = evaluator.eval(injected)
        # logging.debug(f'\"{cond}\" evaluated to: {result}')

        return result

    def resolve(self, config, scope: Mapping):
        """
        Resolves the stored configuration into literal terms at runtime
        """
        result = config

        if isinstance(config, list):
            result = []
            for _, v in enumerate(config):
                parsed = self.resolve(v, scope)
                if not v or parsed is not None:
                    result.append(parsed)
        elif isinstance(config, str):
            result = self._inject_config_str(config, scope)

        return result


class LazyDynamicBase:
    """
    Base object that allows lazy resolution of configuration data
    """

    def __init__(
        self,
        data=None,
        resolver: DataResolver = None,
        scope: DataResolverScope = None,
        parent=None,
    ):
        self._data = data
        self._resolver = resolver
        self._scope = scope
        self._parent = parent
        self._transform_map = {list: LazyDynamicSequence, dict: LazyDynamicDotMap}

    def _transform_object(self, data):
        eval_locals = self._resolver.build_eval_locals(self, self._scope)
        for k, v in self._transform_map.items():
            if isinstance(data, k):
                resolved = v(data, self._resolver, self._scope, self)
                return resolved
        return self._resolver.resolve(data, eval_locals)


class LazyDynamicSequence(LazyDynamicBase, Sequence):
    """
    Lazy dynamic implementation of Sequence.
    """

    def __init__(
        self,
        data: Sequence = [],
        resolver: DataResolver = None,
        scope: DataResolverScope = None,
        parent=None,
    ):
        super().__init__(data, resolver, scope, parent)

    def __getitem__(self, key):
        return self._transform_object(self._data[key])

    def __len__(self):
        return len(self._data)

    def with_scope(self, scope: DataResolverScope):
        return LazyDynamicSequence(self._data, self._resolver, scope)


class LazyDynamicMapping(LazyDynamicBase, Mapping):
    """
    Lazy dynamic implementation of Mapping.
    """

    def __init__(
        self,
        data: Mapping = {},
        resolver: DataResolver = None,
        scope: DataResolverScope = None,
        parent=None,
    ):
        super().__init__(data, resolver, scope, parent)
        self._expressions = self._data.get("@expressions", {})
        self._conditions = self._data.get("@conditions", {})

        # strip special members and members that don't match conditions immediately
        self._data = {k: v for k, v in self._data.items() if self._eval_condition(k)}

    def _eval_condition(self, key: str):
        if key in {"@expressions", "@conditions"}:
            return False
        condition = self._conditions.get(key)
        if isinstance(condition, str) and not self._resolver.eval(
            condition, self, self._scope
        ):
            return False
        return True

    def _transform_kv(self, key: str, value):
        """
        Evaluates @expressions and @conditions on a key
        """
        expression = self._expressions.get(key)
        if isinstance(expression, str):
            value = self._resolver.eval(expression, self, self._scope)

        return self._transform_object(value)

    def __getitem__(self, key):
        return self._transform_kv(key, self._data.get(key))

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def get(self, key, default=None):
        result = self._data.get(key, default)
        if result is None:
            return None

        return self._transform_kv(key, result)

    def with_scope(self, scope: DataResolverScope):
        return LazyDynamicMapping(self._data, self._resolver, scope)


class LazyDynamicDotMap(LazyDynamicMapping):
    """
    Lazy dynamic implementation of DotMap.
    """

    def __init__(
        self,
        data: Mapping = {},
        resolver: DataResolver = None,
        scope: DataResolverScope = None,
        parent=None,
    ):
        super().__init__(data, resolver, scope, parent)

        dmap = DotMap()
        dmap._map = self._data
        self._data = dmap

    def __getattr__(self, k):
        if k in {
            "_data",
            "_resolver",
            "_scope",
            "_transform_map",
            "_expressions",
            "_conditionals",
            "_dotmap",
        }:
            return super(self.__class__, self).__getattribute__(k)
        return self._transform_kv(k, self._data.__getattr__(k))

    def with_scope(self, scope):
        return LazyDynamicDotMap(self._data, self._resolver, scope)


class ConfigurationUtilities:
    @staticmethod
    def parse_root_config(root: Path, config: dict) -> dict:
        # validate the root schema
        schema_path = Path(cas.__file__).parent.absolute().joinpath("schemas")
        with open(schema_path.joinpath("root.json"), "r") as f:
            root_schema = json.load(f)
        DefaultValidatingDraft7Validator(root_schema).validate(config)

        # validate all subsystem options
        validators = {}
        prefix = "cas.subsystems."
        for k, subsystem in config["subsystems"].items():
            if k in {"@conditions", "@expressions"}:
                continue

            module = subsystem.module

            # validating third-party subsystems is not supported yet
            if not module.startswith(prefix):
                continue
            sub_name = module[len(prefix) :]

            if sub_name not in validators:
                sub_path = schema_path.joinpath("subsystems", f"{sub_name}.json")
                if not sub_path.exists():
                    raise Exception(f"unable to find schema for subsystem '{sub_name}'")
                with open(sub_path, "r") as f:
                    validators[sub_name] = DefaultValidatingDraft7Validator(
                        json.load(f)
                    )
            if subsystem.get("options") is None:
                continue
            validators[sub_name].validate(subsystem.options)

        # setup the dotmap that we'll use to perform lazy resolution
        config = DotMap(config)

        config.path.root = root
        config.path.content = root.joinpath("content")
        config.path.game = root.joinpath("game")
        config.path.src = root.joinpath("src")

        config.path.devtools = config.path.src.joinpath("devtools")
        config.path.secrets = config.path.devtools.joinpath("buildsys", "secrets")
        config.path.vproject = config.path.game.joinpath(
            config.options.project
        ).resolve()

        # create the root resolver and the map
        resolver = DataResolver(config)
        return LazyDynamicDotMap(config.toDict(), resolver)



from cas.common.models import BuildResult, BuildSubsystem


class CustomSubsystem(BuildSubsystem):
    """
    Subsystem that invokes a custom program on build and clean steps.
    """

    def build(self) -> BuildResult:
        args = self.config.get("build")
        return True if not args else self.env.run_tool(args, cwd=self.config.cwd) == 0

    def clean(self) -> bool:
        args = self.config.get("clean")
        return True if not args else self.env.run_tool(args, cwd=self.config.cwd) == 0


_subsystem = CustomSubsystem



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="cas",
    version="1.0.6",
    description="Chaos Automation System",
    url="https://github.com/ChaosInitiative/CAS",
    author="Chaos Initiative",
    author_email="contact@chaosinitiative.com",
    license="mit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(include=["cas", "cas.*"]),
    package_data={"cas": ["schemas"]},
    entry_points={"console_scripts": ["casbuild = cas.cli:main"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    install_requires=[
        "jsonschema",
        "simpleeval",
        "appdirs",
        "dotmap",
        "tqdm",
        "vdf",
        "requests",
    ],
    dependency_links=["https://github.com/TeamSpen210/srctools#egg=srctools"],
)



from cas.common.assets.models import (
    Asset,
    AssetBuildContext,
    BatchedDriver,
    PrecompileResult,
)

from typing import List

EXT_MAP = {"kv": "ekv", "nut": "nuc"}


class ViceDriver(BatchedDriver):
    """
    Driver that encrypts files with an ICE key using VICE
    """

    def _tool_name(self):
        return "vice"

    def precompile(self, context: AssetBuildContext, asset: Asset) -> List[str]:
        asset.outpath = asset.path.with_suffix(".ekv")
        return PrecompileResult([asset.path], [asset.outpath])

    def compile_all(self, context: AssetBuildContext, assets: List[Asset]) -> bool:
        key = context.config.options.key
        if len(key) != 8:
            raise Exception("ICE key must be exactly 8 characters long")

        args = [str(self.tool), "-quiet", "-nopause", "-encrypt", key, "-newext", "ekv"]
        for asset in assets:
            args.append(str(asset.path))

        result = self.env.run_tool(args)
        return result == 0


_driver = ViceDriver



import os
import json
import argparse
import logging
import multiprocessing
from pathlib import Path
from dotmap import DotMap

import cas.common.utilities as utilities
from cas.common.sequencer import Sequencer


def _resolve_root_path() -> Path:
    root = Path.cwd()
    while True:
        dirs = os.listdir(root)
        if "content" in dirs and "game" in dirs:
            # we may have a match, verify we have a valid path
            if Path(os.path.join(root, "content", "cas.json")).exists():
                return Path(root).resolve()
        if not root.parent or root.parent == root:
            break
        root = root.parent
    return None


def main():
    parser = argparse.ArgumentParser(description="Chaos Automation System CLI")

    parser.add_argument(
        "-p",
        "--path",
        type=str,
        help="Path to the root project folder. Should contain game and content subfolders. If this is not specified, it is autodetected.",
    )
    parser.add_argument(
        "-f", "--force", action="store_true", help="Forces a rebuild of all assets"
    )
    parser.add_argument(
        "-t",
        "--threads",
        type=int,
        default=multiprocessing.cpu_count(),
        help="Number of threads to use to build assets. Defaults to the number of CPUs on the system.",
    )
    parser.add_argument(
        "-c", "--clean", action="store_true", help="Cleans the build environment."
    )
    parser.add_argument(
        "-b",
        "--build-type",
        type=str.lower,
        choices=["trunk", "staging", "release"],
        default="trunk",
        help="The type of the build.",
    )
    parser.add_argument(
        "-s",
        "--build-categories",
        type=str.lower,
        help="Comma-seperated list of categories to include in the build. If not specified, all categories will be assumed.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Write verbose output for debugging.",
    )
    parser.add_argument(
        "-d",
        "--dry-run",
        action="store_true",
        help="Prints the arguments that would be passed to VPC and exits.",
    )

    parser.add_argument(
        "-o",
        "--override",
        action="append",
        help="Overrides the configuration path specified by x.",
    )

    parser.add_argument(
        "-i",
        "--include-subsystems",
        type=str.lower,
        help="Comma-seperated list of subsystems to include in the build.",
    )
    parser.add_argument(
        "-x",
        "--exclude-subsystems",
        type=str.lower,
        help="Comma-seperated list of subsystems to exclude from the build.",
    )

    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    logger = logging.getLogger(__name__)

    # if the root path is not specified, try to autodetect it
    if args.path:
        root_path = Path(args.path).resolve()
    else:
        root_path = _resolve_root_path()
    if not root_path:
        logger.error(
            "Couldn't find the root path. Ensure you're running CAS from within a subdirectory of the root folder, or use the -p/--path argument to specify it."
        )
        exit(1)

    content_path = root_path.joinpath("content")

    config_file = content_path.joinpath("cas.json")
    if not config_file.exists():
        logger.error(
            "Couldn't find cas.json. If you don't yet have one, please see the documentation for a template."
        )
        exit(1)

    with config_file.open("rb") as f:
        config = DotMap(json.load(f))

    cache_path = content_path.joinpath(".cas_cache.json")
    if args.force and cache_path.exists():
        cache_path.unlink()

    # apply overrides
    if not args.override:
        args.override = []
    for x in args.override:
        assert x.count("=") <= 1, "invalid key-value operator"
        if "=" in x:
            spl = x.split("=", 1)
            val = spl[1]

            # if it starts with [ or {, parse as json
            if val.startswith("[") or val.startswith("{"):
                val = json.loads(val)
            elif val.isdigit():
                val = int(val)

            utilities.set_dot_notation(config, spl[0], val)
        else:
            utilities.set_dot_notation(config, x, True)

    config["args"] = {**config.get("args", {}), **vars(args)}
    config["args"]["cli"] = True

    sequencer = Sequencer(root_path, config)
    if not sequencer.run():
        exit(1)



from cas.common.models import BuildResult, BuildSubsystem
from typing import List
from pathlib import Path

import cas.common.utilities as utilities

import os
import hashlib

import vdf


class VPKArchive:
    def __init__(
        self,
        sys: BuildSubsystem,
        prefix: str,
        input_path: Path,
        output_path: Path,
        files: List[Path],
    ):
        self.sys = sys
        self.prefix = prefix
        self.input_path = input_path
        self.output_path = output_path
        self.files = files

    def _md5_file(self, path: str):
        hash = hashlib.md5()
        with open(path, "rb") as f:
            while True:
                data = f.read(65536)
                if not data:
                    break
                hash.update(data)
        return hash.hexdigest()

    def _gen_control_file(self, output: Path, files: List[Path]):
        entries = {}
        for f in files:
            # Ensure the control file itself and VPKs are excluded
            if (
                f.name == output.name
                or f.name == f"{output.name}.bak"
                or f.suffix == ".vpk"
            ):
                continue

            rel = str(os.path.relpath(f, self.input_path)).replace("\\", "/")
            entries[rel] = {"destpath": rel, "md5": self._md5_file(str(f))}

        # VPK needs the files to stay in the same order
        res = dict(sorted(entries.items()))
        with output.open("w") as f:
            vdf.dump(res, f, pretty=True)

    def pack(self) -> bool:
        vpk_path = self.output_path.joinpath(f"{self.prefix}_dir.vpk")
        cfile_path = self.output_path.joinpath(f"control_{self.prefix}.vdf")
        self._gen_control_file(cfile_path, self.files)

        args = [str(self.sys.env.get_tool("vpk")), "-M", "-P"]
        keypair = self.sys.config.get("keypair")
        if keypair:
            args.extend(
                [
                    "-K",
                    keypair["private"].replace("\\", "/"),
                    "-k",
                    keypair["public"].replace("\\", "/"),
                ]
            )
        args.extend(["k", str(vpk_path), str(cfile_path)])

        returncode = self.sys.env.run_tool(args, cwd=self.input_path)
        if returncode != 0:
            return False

        # handle the previous backup file
        bakfile = cfile_path.with_suffix(".vdf.bak")
        if os.path.exists(bakfile):
            os.remove(bakfile)
        os.rename(cfile_path, bakfile)
        return True


class VPKBuildSubsystem(BuildSubsystem):
    """
    Subsystem that packs one or more files into a VPK
    """

    def _get_vpk(self, config: dict) -> VPKArchive:
        prefix = config.prefix

        files = set()
        input_path = Path(config.input).resolve()
        output_path = config.get("output")
        if output_path:
            output_path = Path(output_path).resolve()
        else:
            output_path = input_path

        assert input_path.exists()
        assert output_path.exists()

        files = utilities.rglob_multi(input_path, config.get("files", []))
        if len(files) == 0:
            self._logger.warning("No files to pack!")
            return

        return VPKArchive(self, prefix, input_path, output_path, files)

    def build(self) -> BuildResult:
        outputs = {"files": []}
        for f in self.config.packfiles:
            vpk = self._get_vpk(f)
            pakid = f"{vpk.output_path.parts[-1]}/{vpk.prefix}"
            self._logger.info(f"Packing {len(vpk.files)} files into {pakid}")
            if not vpk.pack():
                self._logger.error(f"Failed to pack {pakid}!")
                return BuildResult(False)

            # capture the file patterns so we can pass them to other subsystems later
            for entry in f.get("files", []):
                # need to move the ! to the start of the pattern, since we're joining!
                exclude = entry.startswith("!")
                if exclude:
                    entry = entry[1:]
                entry = os.path.join(vpk.input_path, entry)
                if exclude:
                    entry = "!" + entry
                outputs["files"].append(entry)

        return BuildResult(True, outputs)

    def clean(self) -> bool:
        for vpk in self.config.packfiles:
            # dupe code here, fix
            input_path = Path(vpk["input"]).resolve()
            output_path = vpk.get("output")
            if output_path:
                output_path = Path(output_path).resolve()
            else:
                output_path = input_path

            for path in output_path.rglob(vpk["prefix"] + "*.vpk"):
                path.unlink()

            ctl = output_path.joinpath("control_" + vpk["prefix"] + ".vdf")
            if ctl.exists():
                ctl.unlink()

            ctl = ctl.with_suffix(".vdf.bak")
            if ctl.exists():
                ctl.unlink()

        return True


_subsystem = VPKBuildSubsystem
